@@ -28,8 +28,7 @@ pub(crate) struct ConnectionInner {
 pub(crate) enum SessionState {
     Opening(Option<oneshot::Sender<Session>>, Cell<ConnectionInner>),
     Established(Cell<SessionInner>),
-    #[allow(dead_code)]
-    Closing(Option<oneshot::Sender<Result<(), AmqpProtocolError>>>),
+    Closing(Cell<SessionInner>),
 }
 
 impl SessionState {
@@ -95,20 +94,25 @@ impl Connection {
 
     /// Gracefully close connection
     pub fn close(&self) -> impl Future<Output = Result<(), AmqpProtocolError>> {
-        self.0.get_ref().io.close();
+        let inner = self.0.get_mut();
+        inner.post_frame(AmqpFrame::new(0, Frame::Close(Close { error: None })));
+        inner.io.close();
         Ready::Ok(())
     }
 
-    // TODO: implement
     /// Close connection with error
-    pub fn close_with_error<E>(
-        &self,
-        _err: &E,
-    ) -> impl Future<Output = Result<(), AmqpProtocolError>>
+    pub fn close_with_error<E>(&self, err: E) -> impl Future<Output = Result<(), AmqpProtocolError>>
     where
         Error: From<E>,
     {
-        self.0.get_ref().io.close();
+        let inner = self.0.get_mut();
+        inner.post_frame(AmqpFrame::new(
+            0,
+            Frame::Close(Close {
+                error: Some(err.into()),
+            }),
+        ));
+        inner.io.close();
         Ready::Ok(())
     }
 
@@ -153,6 +157,14 @@ impl Connection {
         }
     }
 
+    pub(crate) fn close_session(&self, id: usize) {
+        if let Some(state) = self.0.get_mut().sessions.get_mut(id) {
+            if let SessionState::Established(inner) = state {
+                *state = SessionState::Closing(inner.clone());
+            }
+        }
+    }
+
     pub(crate) fn post_frame(&self, frame: AmqpFrame) {
         #[cfg(feature = "frame-trace")]
         log::trace!("outgoing: {:#?}", frame);
@@ -196,6 +208,9 @@ impl ConnectionInner {
     }
 
     pub(crate) fn post_frame(&mut self, frame: AmqpFrame) {
+        #[cfg(feature = "frame-trace")]
+        log::trace!("outgoing: {:#?}", frame);
+
         if let Err(e) = self.io.write().encode(frame, &self.codec) {
             self.set_error(e.into());
         }
@@ -295,18 +310,19 @@ impl ConnectionInner {
                 return Ok(Action::None);
             }
             Frame::Close(close) => {
-                self.set_error(AmqpProtocolError::Closed(close.error.clone()));
-
                 if self.state == ConnectionState::Closing {
                     log::trace!("Connection closed: {:?}", close);
                     self.set_error(AmqpProtocolError::Disconnected);
+                    return Ok(Action::None);
                 } else {
                     log::trace!("Connection closed remotely: {:?}", close);
+                    let err = AmqpProtocolError::Closed(close.error.clone());
+                    self.set_error(err.clone());
                     let close = Close { error: None };
                     self.post_frame(AmqpFrame::new(0, close.into()));
                     self.state = ConnectionState::RemoteClose;
+                    return Ok(Action::RemoteClose(err));
                 }
-                return Ok(Action::None);
             }
             Frame::Begin(begin) => {
                 // response Begin for open session
@@ -382,16 +398,16 @@ impl ConnectionInner {
                 }
                 _ => session.get_mut().handle_frame(frame),
             },
-            SessionState::Closing(ref mut tx) => match frame {
+            SessionState::Closing(ref mut session) => match frame {
                 Frame::End(frm) => {
                     trace!("Session end is confirmed: {:?}", frm);
-                    if let Some(tx) = tx.take() {
-                        let _ = tx.send(Ok(()));
-                    }
+                    let action = session
+                        .get_mut()
+                        .end(AmqpProtocolError::SessionEnded(frm.error));
                     if let Some(token) = self.sessions_map.remove(&channel_id) {
                         self.sessions.remove(token);
                     }
-                    Ok(Action::None)
+                    Ok(action)
                 }
                 frm => {
                     trace!("Got frame after initiated session end: {:?}", frm);
